{#
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
#}
{% extends "airflow/master.html" %}

{% block content %}
    {{ super() }}
    <h2>{{ title }}</h2>
  {% set version_label = 'Version' %}
    {% if airflow_version %}
        <h4>{{ version_label }} : <a href="https://pypi.python.org/pypi/apache-airflow/{{ airflow_version }}">{{ airflow_version }}</a></h4>
    {% else %}
        <h4>{{ version_label }} : Not Available</h4>
    {% endif %}
  <h4>Git Version :{% if git_version %} {{ git_version }} {% else %} Not Available {% endif %}</h4>
    <hr>

{% endblock %}
