{# This is a dummy template, we can use that template to add attributes unique
   to discreteBarChart #}

{% extends "content.html" %}
{% block body %}

    {% block data %}
        {{super()}}
    {% endblock data %}

    {% block init %}
        {{super()}}
    {% endblock init %}

    {% block axes %}
        {{super()}}
    {% endblock axes %}
    
    {% block custoattr %}
    {{super()}}
    {% endblock custoattr %}

    {% block inject %}
    {{ super() }}
    {% endblock inject %}
    
    {% block close %}
    {{ super() }}
    {% endblock close %}

{% endblock body %}
