{#
 Licensed to the Apache Software Foundation (ASF) under one
 or more contributor license agreements.  See the NOTICE file
 distributed with this work for additional information
 regarding copyright ownership.  The ASF licenses this file
 to you under the Apache License, Version 2.0 (the
 "License"); you may not use this file except in compliance
 with the License.  You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing,
 software distributed under the License is distributed on an
 "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 KIND, either express or implied.  See the License for the
 specific language governing permissions and limitations
 under the License.
#}

<!DOCTYPE html>
<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html;charset=utf-8"/>
    <title>{{ chart.label }}</title>
    <link rel="stylesheet" type="text/css" href="/static/para/parallel.css" />
  </head>
  <body class="dark">
  <div id="header">
      <h1>{{ chart.label }}</h1>
    <button title="Zoom in on selected data" id="keep-data" disabled="disabled">Keep</button>
    <button title="Remove selected data" id="exclude-data" disabled="disabled">Exclude</button>
    <button title="Export data as CSV" id="export-data">Export</button>
    Color based on: <select id="fields"></select>
    <div class="controls">
      <strong id="selected-count"></strong><!--<strong id="data-count"></strong>-->
      <div class="fillbar"><div id="selected-bar"><div id="rendered-bar">&nbsp;</div></div></div>
      Lines at <strong id="opacity"></strong> opacity.
      <span class="settings">
        <button id="hide-ticks">Hide Ticks</button>
        <button id="show-ticks" disabled="disabled">Show Ticks</button>
        <button id="dark-theme" disabled="disabled">Dark</button>
        <button id="light-theme">Light</button>
      </span>
    </div>
    <div style="clear:both;"></div>
  </div>
  <div id="chart">
    <canvas id="background"></canvas>
    <canvas id="foreground"></canvas>
    <canvas id="highlight"></canvas>
    <svg></svg>
  </div>
  <div id="wrap">
    <div class="third" id="controls">

      <div class="little-box">
        <h3>Controls</h3>
        <p>
          <strong>Brush</strong>: Drag vertically along an axis.<br/>
          <strong>Remove Brush</strong>: Tap the axis background.<br/>
          <strong>Reorder Axes</strong>: Drag a label horizontally.<br/>
          <strong>Invert Axis</strong>: Tap an axis label.<br/>
          <strong>Remove Axis</strong>: Drag axis label to the left edge.<br/>
        </p>
      </div>
      <div class="little-box">
        <h3>Credits &amp; License</h3>
          <p>
            <a href="https://github.com/gameclosure/webgl-2d">Webgl-2d</a>
          </p>
          <p>
          Adapted from examples by<br/>
          <a href="http://bl.ocks.org/1341021">Mike Bostock</a> and <a href="http://bl.ocks.org/1341281">Jason Davies</a><br/>
          </p>
          <p>
            Copyright &copy; 2012, Kai Chang<br/>
            All rights reserved. Released under the <a href="http://opensource.org/licenses/bsd-3-clause">BSD License</a>.
          </p>
        </p>
      </div>
    </div>
    <div  class="third">
      <small>
        <!--Last rendered <strong id="render-speed"></strong> lines-->
      </small>
      <h3>Categories</h3>
      <p id="legend">
      </p>
    </div>
    <div class="third">
      <h3>Sample of 25 entries <input type="text" id="search" placeholder="Search..."></input></h3>
      <p id="food-list">
      </p>
    </div>
  </div>
  </body>
  <script src="/static/d3.v3.min.js"></script>
  <script src="/static/underscore.js"></script>
  <script src="/static/para/webgl-2d.js"></script>
  <script src="/static/para/parallel.js"></script>
</html>
